/*---------------------------------------------------------------------------*\
| 
| Developer: alexander.vakhrushev@unileoben.ac.at
|
| https://sites.google.com/site/smmp01/home/staff-member/dr-a-vakhrushev
|
/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2010 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

Application
    icoBlockedCellFoam

Description
    Transient solver for incompressible, laminar flow of Newtonian fluids.
    Excluding the flow calculation for the cells in the predefined location
    is implemented.

\*---------------------------------------------------------------------------*/

#include "fvCFD.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

int main(int argc, char *argv[])
{
    #include "setRootCase.H"

    #include "createTime.H"
    #include "createMesh.H"
    #include "createFields.H"
    #include "initContinuityErrs.H"

    // * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

    Info<< "\nStarting time loop\n" << endl;

    //  Find the cell at the defined location to exclude from calculation
    label cell_ = mesh.findCell(position_);
    Info<< "Location " << position_ << " is at the cell #"
        <<  cell_ << "." << nl << endl;

    while (runTime.loop())
    {
        Info<< "Time = " << runTime.timeName() << nl << endl;

        #include "readPISOControls.H"
        #include "CourantNo.H"

        fvVectorMatrix UEqn
        (
            fvm::ddt(U)
          + fvm::div(phi, U)
          - fvm::laplacian(nu, U)
        );

        //  Modify matrix coefficients for momentum equation
        {
            //  Access lower, upper triangles and diagonal to modify
            scalarField& lower_ = UEqn.lower();
            scalarField& upper_ = UEqn.upper();
            scalarField& diag_ = UEqn.diag();

            //  Get matrix LDU addressing.
            //  Connects face indexing for the off-diagonal elements
            //  with cell indexing numbering of the diagonal
            const unallocLabelList& l_ = UEqn.lduAddr().lowerAddr();
            const unallocLabelList& u_ = UEqn.lduAddr().upperAddr();

            //  Set diagonal coefficient to identity value
            diag_[cell_] = scalar(1);

            //  Loop over all off-diagonal elements.
            //  Set them to zero value to exclude the influence
            //  of the neighbouring cells
            for (register label face_=0; face_<l_.size(); face_++)
            {
                if (l_[face_] == cell_)
                {
                    lower_[face_] = scalar(0);
                }

                if (u_[face_] == cell_)
                {
                    upper_[face_] = scalar(0);
                }
            }

            //  Exclude source term, which comes from dU/dt discretization
            UEqn.source()[cell_] = vector(0, 0, 0);
//            Info<< "UEqn sorce term:" << nl
//                << UEqn.source() << nl << endl;

            //  Copy matrix to manipulate the pressure gradient source term.
            //  Initial UEqn is required for pressure correction.
            fvVectorMatrix UpGradEqn = UEqn;

            //  Include pressure for all cells
            UpGradEqn += fvc::grad(p);

            //  Exclude any sources for the cell in desired location
            UpGradEqn.source()[cell_] = vector(0, 0, 0);

            //  Solve corrected momentum predictor equation
            UpGradEqn.solve();
        }

        // --- PISO loop

        for (int corr=0; corr<nCorr; corr++)
        {
            volScalarField rUA = 1.0/UEqn.A();

            U = rUA*UEqn.H();
            phi = (fvc::interpolate(U) & mesh.Sf())
                + fvc::ddtPhiCorr(rUA, U, phi);

            adjustPhi(phi, U, p);

            for (int nonOrth=0; nonOrth<=nNonOrthCorr; nonOrth++)
            {
                fvScalarMatrix pEqn
                (
                    fvm::laplacian(rUA, p) == fvc::div(phi)
                );

                pEqn.setReference(pRefCell, pRefValue);

                //  Modify pressure equation matrix
                {
                    //  Take care about matrix symmetry.
                    //  Due to read-write access to the off-diagonal elements
                    //  it can be converted to assymetric one!
                    //  Thus either lower or upper elements are used.
                    scalarField& off_diag_ =
                        (pEqn.hasUpper()) 
                            ? pEqn.upper() 
                            : pEqn.lower();

                    //  Get diagonal coefficients
                    scalarField& diag_ = pEqn.diag();

                    //  LDU off-diagonal addressing for symmetric matrix
                    const unallocLabelList& o_ =
                        (pEqn.hasUpper())
                            ? pEqn.lduAddr().upperAddr()
                            : pEqn.lduAddr().lowerAddr();

                    //  Set diagonal to identity
                    diag_[cell_] = scalar(1);

                    //  Remove the influence of the neighbours 
                    for (register label face_=0; face_<o_.size(); face_++)
                    {
                        if (o_[face_] == cell_)
                        {
                            off_diag_[face_] = scalar(0);
                        }
                    }

                    //  Remove source for desired cell
                    pEqn.source()[cell_] = scalar(0);
                }

                pEqn.solve();

                if (nonOrth == nNonOrthCorr)
                {
                    phi -= pEqn.flux();
                }
            }

            #include "continuityErrs.H"

            U -= rUA*fvc::grad(p);
            U.correctBoundaryConditions();
        }


        runTime.write();

        Info<< "ExecutionTime = " << runTime.elapsedCpuTime() << " s"
            << "  ClockTime = " << runTime.elapsedClockTime() << " s"
            << nl << endl;
    }

    Info<< "End\n" << endl;

    return 0;
}


// ************************************************************************* //
